#!/usr/bin/env python
'''Class for plotting simulation data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

# Base python imports
import numpy as np
import os
import scipy.stats
import scipy.signal as signal
import warnings
import verdict

import matplotlib
matplotlib.use('PDF')
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as plt_colors
import matplotlib.gridspec as gridspec
import matplotlib.patheffects as path_effects
import matplotlib.transforms as transforms

import galaxy_dive.utils.mp_utils as mp_utils
import galaxy_dive.utils.utilities as utilities

import galaxy_dive.plot_data.plotting as gen_plot
import galaxy_dive.plot_data.pu_colormaps as pu_cm

########################################################################
########################################################################

class GenericPlotter( object ):

    @utilities.store_parameters
    def __init__( self, data_object, label=None, color='black', ):
        '''
        Args:
            data_object ( generic_data.GenericData object or subclass of such ) : The data container to use.
        '''

        pass

    ########################################################################
    # Alternate inherent methods
    ########################################################################

    def __getattr__( self, attr):
        '''By replacing getattr with the following code, we allow automatically searching the data_object
        for the appropriate attribute as well, while losing none of the original functionality.
        '''

        print( "Attribute {} not found in plotting object. Checking data object.".format( attr ) )

        return getattr( self.data_object, attr )

    ########################################################################
    # Specific Generic Plots
    ########################################################################

    def histogram( self,
        data_key,
        provided_data = None,
        provided_hist = None,
        weight_key = None,
        slices = None,
        ax = None,
        fix_invalid = False,
        mask_zeros = False,
        invalid_fix_method = None,
        bins = 32,
        normed = True,
        norm_type = 'probability',
        scaling = None,
        smooth = False,
        smoothing_window_length = 9,
        smoothing_polyorder = 3,
        histogram_style = 'step',
        color = 'black',
        linestyle = '-',
        linewidth = 3.5,
        alpha = 1.,
        x_range = None, y_range = None,
        x_label = None, y_label = None,
        add_x_label = True, add_y_label = True,
        add_plot_label = True,
        plot_label = None,
        line_label = None,
        label_fontsize = 24,
        x_scale = 'linear', y_scale = 'linear',
        cdf = False,
        vertical_line = None,
        vertical_line_kwargs = { 'linestyle': '--', 'linewidth': 3, 'color': 'k', },
        return_dist = False,
        assert_contains_all_data = True,
        data_kwargs = {},
        *args, **kwargs ):
        '''Make a histogram of the data. Extra arguments are passed to
        self.data_object.get_selected_data.

        Args:
            data_key (str) :
                Data key to plot.

            weight_key (str) :
                Data key for data to use as a weight. By None, no weight.

            slices (int or tuple of slices) :
                How to slices the data.

            ax (axis) :
                What axis to use. By default creates a figure and places the axis on it.

            fix_invalid (bool) :
                Throw away invalid values?

            invalid_fix_method (float or int) :
                How to handle invalid values. By None throw them away. Providing a value to this argument instead replaces
                them with that value.

            bins (int or array-like) :
                bins argument to be passed to np.histogram

            normed (bool) :
                Normalize the histogram?

            color (str) :
                Color of histogram.

            linestyle (str) :
                Linestyle of histogram.

            linewidth (float) :
                Linewidth for histogram.

            alpha (float) :
                Alpha value for histogram

            x_range, y_range ( [min,max] ) :
                What are the minimum and maximum x- and y- values to include?
                Defaults to matplotlib's automatic choices

            x_label, ylabel (str) :
                Axes labels. Defaults to the data_key for the x-axis and "Normalized Histogram" for the y-axis.

            add_x_label, add_y_label (bool) :
                Include axes labels?

            plot_label (str or dict) :
                What to label the plot with. By None, uses self.label.

            line_label (str) :
                What label to give the line.

            label_fontsize (int) :
                Fontsize for the labels.

            x_scale, y_scale (str) :
                What scales to use for the x and y axes.

            cdf (bool) :
                Plot a CDF instead.

            vertical_line (float) :
                Plot a vertical line at this value on the x-axis, if true.

            vertical_line_kwargs (dict) :
                Arguments to pass to ax.plot( for the vertical line.

            return_dist (bool) :
                If True, return the data values and the edges for the histogram.

            assert_contains_all_data (bool) :
                If True, make sure that the histogram plots all selected data.

            *args, **kwargs :
                Extra arguments to pass to self.data_object.get_selected_data()
        '''

        print( "Plotting histogram for {}".format( data_key ) )

        if provided_hist is None:

            if isinstance( slices, int ):
                sl = ( slice(None), slices )
            else:
                sl = slices

            data_kwargs = utilities.merge_two_dicts( data_kwargs, kwargs )

            if provided_data is None:
                data = self.data_object.get_selected_data(
                    data_key,
                    sl=sl,
                    *args, **data_kwargs
                ).copy()
            else:
                data = provided_data.copy()

            if weight_key is None:
                weights = None
            else:

                if 'scale_key' in kwargs:
                    warnings.warn(
                        "Scaling weights by {}. Is this correct?".format(
                            kwargs['scale_key']
                        )
                    )
                weights = self.data_object.get_selected_data( weight_key, sl=sl, *args, **kwargs )

            if fix_invalid:
                if invalid_fix_method is None:
                    data = np.ma.fix_invalid( data ).compressed()
                else:
                    data = np.ma.fix_invalid( data )
                    data.fill_value = invalid_fix_method
                    data = data.filled()

            # Make the histogram itself
            hist, edges = np.histogram( data, bins=bins, weights=weights )

            # Make sure we have all the data in the histogram
            if assert_contains_all_data:
                assert data.size == hist.sum()

            if normed:

                if norm_type == 'probability':
                    hist = hist.astype( float ) / ( hist.sum()*(edges[1] - edges[0]) )
                elif norm_type == 'bin_width':
                    hist = hist.astype( float ) / (edges[1] - edges[0])
                elif norm_type == 'outer_edge':
                    hist = hist.astype( float ) / hist[-1]
                elif norm_type == 'max_value':
                    hist = hist.astype( float ) / hist.max()
                else:
                    raise Exception(
                        "Unrecognized norm_type, {}".format( norm_type )
                    )

            if scaling is not None:
                hist *= scaling

            if cdf:
                hist = np.cumsum( hist )*(edges[1] - edges[0])

        else:
            hist = provided_hist
            edges = bins

        if mask_zeros:
            hist = np.ma.masked_where(
                hist < 1e-14,
                hist,
            )

        if smooth:
            hist = signal.savgol_filter(
                hist,
                window_length = smoothing_window_length,
                polyorder = smoothing_polyorder,
            )


        if ax is None:
            fig = plt.figure( figsize=(11,5), facecolor='white', )
            ax = plt.gca()
        if line_label is None:
            line_label = self.label

        if color is None:
            color = self.color

        # Inserting a 0 at the beginning allows plotting a numpy histogram with a step plot
        if histogram_style == 'step':
            ax.step(
                edges,
                np.insert(hist, 0, 0.),
                color=color,
                linestyle=linestyle,
                linewidth=linewidth,
                label=line_label,
                alpha=alpha,
            )
        elif histogram_style == 'line':
            x_values = 0.5 * ( edges[:-1] + edges[1:] )
            ax.plot(
                x_values,
                hist,
                color=color,
                linestyle=linestyle,
                linewidth=linewidth,
                label=line_label,
                alpha=alpha,
            )
        else:
            raise KeyError(
                "Unrecognized histogram_style, {}".format( histogram_style )
            )

        # Plot a vertical line?
        if vertical_line is not None:
            trans = transforms.blended_transform_factory( ax.transData, ax.transAxes )
            ax.plot(
                [ vertical_line, ]*2,
                [ 0., 1., ],
                transform = trans,
                **vertical_line_kwargs
)

        # Plot label
        if add_plot_label:
            if plot_label is None:
                plt_label = ax.annotate(
                    s = self.label,
                    xy = (0.,1.0),
                    va = 'bottom',
                    xycoords = 'axes fraction',
                    fontsize = label_fontsize,
                 )
            elif isinstance( plot_label, str ):
                plt_label = ax.annotate(
                    s = plot_label,
                    xy = (0.,1.0),
                    va = 'bottom',
                    xycoords = 'axes fraction',
                    fontsize = label_fontsize,
                 )
            elif isinstance( plot_label, dict ):
                plt_label = ax.annotate( **plot_label )
            elif plot_label is None:
                pass
            else:
                raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )

        # Add axis labels
        if add_x_label:
            if x_label is None:
                x_label = data_key
            ax.set_xlabel( x_label, fontsize=label_fontsize )
        if add_y_label:
            if y_label is None:
                if not cdf:
                    y_label = r'Normalized Histogram'
                else:
                    y_label = r'CDF'
            ax.set_ylabel( y_label, fontsize=label_fontsize )

        if x_range is not None:
            ax.set_xlim( x_range )
        if y_range is not None:
            ax.set_ylim( y_range )

        ax.set_xscale( x_scale )
        ax.set_yscale( y_scale )

        if return_dist:
            return hist, edges

    ########################################################################

    def histogram2d( self,
        x_key, y_key,
        x_data = None, y_data = None,
        weight_key = None,
        x_data_args = {}, y_data_args = {},
        weight_data_args = {},
        slices = None,
        ax = None,
        x_range = None, y_range = None,
        x_scale = 'linear', y_scale = 'linear', z_scale = 'log',
        n_bins = 128,
        average = False,
        normed = False,
        hist_div_arr = None,
        conditional_y = False,
        y_div_function = None,
        vmin = None, vmax = None,
        min_bin_value_displayed = None,
        zorder = 0,
        add_colorbar = True,
        cmap = pu_cm.magma,
        colorbar_args = None,
        x_label = None, y_label = None,
        add_x_label = True, add_y_label = True,
        plot_label = None,
        outline_plot_label = False,
        label_galaxy_cut = False,
        label_redshift = False,
        label_fontsize = 24,
        tick_param_args = None,
        out_dir = None,
        save_file = None,
        close_plot_after_saving = True,
        fix_invalid = True,
        line_slope = None,
        cdf = False,
        horizontal_line = None, vertical_line = None,
        horizontal_line_kwargs = { 'linestyle': '--', 'linewidth': 5, 'color': '#337DB8', },
        vertical_line_kwargs = { 'linestyle': '--', 'linewidth': 5, 'color': '#337DB8', },
        return_dist = False,
        *args, **kwargs ):
        '''Make a 2D histogram of the data. Extra arguments are passed to get_selected_data.

        Args:
            x_key, y_key (str) : Data keys to plot.
            weight_key (str) : Data key for data to use as a weight. By None, no weight.
            x_data_args, y_data_args (dicts) : Keyword arguments to be passed only to x or y.
            slices (int or tuple of slices) : How to slices the data.
            ax (axis) : What axis to use. By None creates a figure and places the axis on it.
            x_range, y_range ( (float, float) ) : Histogram edges. If None, all data is enclosed. If list, set manually.
                If float, is +- x_range*length scale at that snapshot.
            n_bins (int) : Number of bins in the histogram.
            vmin, vmax (float) : Limits for the colorbar.
            aspect (str) : What should the aspect ratio of the plot be?
            plot_halos (bool) : Whether or not to plot merger tree halos on top of the histogram.
                Only makes sense for when dealing with positions.
            add_colorbar (bool) : If True, add a colorbar to colorbar_args
            colorbar_args (axis) : What axis to add the colorbar to. By None, is ax.
            x_label, ylabel (str) : Axes labels.
            add_x_label, add_y_label (bool) : Include axes labels?
            plot_label (str or dict) : What to label the plot with. By None, uses self.label.
                Can also pass a dict of full args.
            outline_plot_label (bool) : If True, add an outline around the plot label.
            label_galaxy_cut (bool) : If true, add a label that indicates how the galaxy was defined.
            label_redshift (bool) : If True, add a label indicating the redshift.
            label_fontsize (int) : Fontsize for the labels.
            tick_param_args (args) : Arguments to pass to ax.tick_params. By None, don't change inherent defaults.
            out_dir (str) : If given, where to save the file.
            fix_invalid (bool) : Fix invalid values.
            line_slope (float) : If given, draw a line with the given slope.
        '''

        if isinstance( slices, int ):
            sl = ( slice(None), slices )
        else:
            sl = slices

        varying_kwargs = {
            'x': x_data_args,
            'y': y_data_args,
            'weight': weight_data_args
        }
        data_kwargs = utilities.dict_from_defaults_and_variations( kwargs, varying_kwargs )

        # Get data
        if x_data is None:
            x_data = self.data_object.get_selected_data( x_key, sl=sl, *args, **data_kwargs['x'] ).copy()
        if y_data is None:
            y_data = self.data_object.get_selected_data( y_key, sl=sl, *args, **data_kwargs['y'] ).copy()

        if y_div_function is not None:
            y_div_values = y_div_function( x_data )
            y_data /= y_div_values

        # Fix NaNs
        if fix_invalid:
            x_mask = np.ma.fix_invalid( x_data ).mask
            y_mask = np.ma.fix_invalid( y_data ).mask
            mask = np.ma.mask_or( x_mask, y_mask )

            x_data = np.ma.masked_array( x_data, mask=mask ).compressed()
            y_data = np.ma.masked_array( y_data, mask=mask ).compressed()

        if weight_key is None:
            weights = None
        else:
            weights = self.data_object.get_selected_data(
                weight_key,
                sl=sl,
                *args,
                **data_kwargs['weight']
            ).flatten()

            if fix_invalid:
                weights = np.ma.masked_array( weights, mask=mask ).compressed()

        if x_range is None:
            x_range = [ x_data.min(), x_data.max() ]
        elif isinstance( x_range, float ):
            x_range = np.array( [ -x_range, x_range ])*self.data_object.length_scale[slices]
        if y_range is None:
            y_range = [ y_data.min(), y_data.max() ]
        elif isinstance( y_range, float ):
            y_range = np.array( [ -y_range, y_range ])*self.data_object.length_scale[slices]

        if x_scale == 'log':
            x_edges = np.logspace( np.log10( x_range[0] ), np.log10( x_range[1] ), n_bins )
        else:
            x_edges = np.linspace( x_range[0], x_range[1], n_bins )
        if y_scale == 'log':
            y_edges = np.logspace( np.log10( y_range[0] ), np.log10( y_range[1] ), n_bins )
        else:
            y_edges = np.linspace( y_range[0], y_range[1], n_bins )

        # Make the histogram
        hist2d, x_edges, y_edges = np.histogram2d( x_data, y_data, [x_edges, y_edges], weights=weights, normed=normed )

        # If doing an average, divide by the number in each bin
        if average:
            average_hist2d, x_edges, y_edges = np.histogram2d( x_data, y_data, [x_edges, y_edges], normed=normed )
            hist2d /= average_hist2d

        # If making the y-axis conditional, divide by the distribution of data for the x-axis.
        if conditional_y:
            hist_x, x_edges = np.histogram( x_data, x_edges, normed=normed )
            hist2d /= hist_x[:,np.newaxis]

        # Divide the histogram bins by this array
        if hist_div_arr is not None:
            hist2d /= hist_div_arr

        # Mask bins below a specified value
        if min_bin_value_displayed is not None:
            hist2d = np.ma.masked_where(
                hist2d < min_bin_value_displayed,
                hist2d,
            )

        # Plot
        if ax is None:
            fig = plt.figure( figsize=(10,9), facecolor='white' )
            ax = plt.gca()

        if z_scale == 'linear':
            norm = plt_colors.Normalize()
        elif z_scale == 'log':
            norm = plt_colors.LogNorm()

        if cdf:
            raise Exception(
                "Not implemented yet. When implementing, use utilities.cumsum2d"
            )

        im = ax.pcolormesh(
            x_edges,
            y_edges,
            hist2d.transpose(),
            cmap = cmap,
            norm = norm,
            vmin = vmin,
            vmax = vmax,
            zorder = zorder,
        )

        # Add a colorbar
        if add_colorbar:
            if colorbar_args is None:
                colorbar_args = ax
                cbar = gen_plot.add_colorbar( colorbar_args, im, method='ax' )
            else:
                colorbar_args['color_object'] = im
                cbar = gen_plot.add_colorbar( **colorbar_args )
            cbar.ax.tick_params( labelsize=20 )

        # Plot Line for easier visual interpretation
        if line_slope is not None:
            line_x = np.array( [ x_data.min(), x_data.max() ] )
            line_y = line_slope*line_x
            ax.plot( line_x, line_y, linewidth=3, linestyle='dashed', )

        if horizontal_line is not None:
            trans = transforms.blended_transform_factory( ax.transAxes, ax.transData )
            ax.plot( [ 0., 1. ], [ horizontal_line, ]*2, transform=trans, **horizontal_line_kwargs )
        if vertical_line is not None:
            trans = transforms.blended_transform_factory( ax.transData, ax.transAxes )
            ax.plot( [ vertical_line, ]*2, [ 0., 1. ], transform=trans, **vertical_line_kwargs )

        # Plot label
        if plot_label is not None:
            if plot_label is None:
                plt_label = ax.annotate(
                    s = self.label,
                    xy = (0.,1.0),
                    va = 'bottom',
                    xycoords = 'axes fraction',
                    fontsize = label_fontsize,
                 )
            elif isinstance( plot_label, str ):
                plt_label = ax.annotate(
                    s = plot_label,
                    xy = (0.,1.0),
                    va = 'bottom',
                    xycoords = 'axes fraction',
                    fontsize = label_fontsize,
                 )
            elif isinstance( plot_label, dict ):
                plt_label = ax.annotate( **plot_label )
            else:
                raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )
            if outline_plot_label:
                plt_label.set_path_effects([ path_effects.Stroke(linewidth=3, foreground='black'), path_effects.Normal() ])

        # Upper right label (info label)
        info_label = ''
        if label_galaxy_cut:
            info_label = r'$r_{ \rm cut } = ' + '{:.3g}'.format( self.data_object.galids.parameters['galaxy_cut'] ) + 'r_{ s}$'
        if label_redshift:
            try:
                info_label = r'$z=' + '{:.3f}'.format( self.data_object.redshift ) + '$'+ info_label
            except ValueError:
                info_label = r'$z=' + '{:.3f}'.format( self.data_object.redshift.values[sl[1]] ) + '$'+ info_label
        if label_galaxy_cut or label_redshift:
            ax.annotate( s=info_label, xy=(1.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,
                ha='right' )

        # Add axis labels
        if add_x_label:
            if x_label is None:
                x_label = x_key
            ax.set_xlabel( x_label, fontsize=label_fontsize )
        if add_y_label:
            if y_label is None:
                y_label = y_key
            ax.set_ylabel( y_label, fontsize=label_fontsize )

        # Limits
        ax.set_xlim( x_range )
        ax.set_ylim( y_range )

        # Scale
        ax.set_xscale( x_scale )
        ax.set_yscale( y_scale )

        # Set tick parameters
        if tick_param_args is not None:
            ax.tick_params( **tick_param_args )

        # Save the file
        if out_dir is not None:
            if save_file is None:
                save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[slices] )

            gen_plot.save_fig( out_dir, save_file, fig=fig, dpi=75 )

            if close_plot_after_saving:
                plt.close()

        # Return?
        if return_dist:
            return hist2d, x_edges, y_edges

    ########################################################################

    def statistic_and_interval(
        self,
        x_key, y_key,
        x_data = None, y_data = None,
        weights = None,
        statistic = 'median',
        lower_percentile = 16,
        upper_percentile = 84,
        plot_interval = True,
        x_data_args = {}, y_data_args = {},
        ax = None,
        slices = None,
        fix_invalid = False,
        bins = 64,
        linewidth = 3,
        linestyle = '-',
        color = 'k',
        label = None,
        zorder = 100,
        alpha = 0.5,
        plot_label = None,
        add_plot_label = True,
        plot_label_kwargs = {
            'xy': (0.,1.0),
            'va': 'bottom',
            'xycoords': 'axes fraction',
            'fontsize': 22,
        },
        return_values = False,
        *args, **kwargs
    ):

        if isinstance( slices, int ):
            sl = ( slice(None), slices )
        else:
            sl = slices

        varying_kwargs = {
            'x': x_data_args,
            'y': y_data_args,
        }
        data_kwargs = utilities.dict_from_defaults_and_variations( kwargs, varying_kwargs )

        # Get data
        if x_data is None:
            x_data = self.data_object.get_selected_data( x_key, sl=sl, *args, **data_kwargs['x'] ).copy()
        if y_data is None:
            y_data = self.data_object.get_selected_data( y_key, sl=sl, *args, **data_kwargs['y'] ).copy()

        # Fix NaNs
        if fix_invalid:
            x_mask = np.ma.fix_invalid( x_data ).mask
            y_mask = np.ma.fix_invalid( y_data ).mask
            mask = np.ma.mask_or( x_mask, y_mask )

            x_data = np.ma.masked_array( x_data, mask=mask ).compressed()
            y_data = np.ma.masked_array( y_data, mask=mask ).compressed()

        # Calculate the statistic
        if statistic == 'weighted_mean':

            assert weights is not None, "Need to provide weights."

            weighted_sum, bin_edges, binnumber = scipy.stats.binned_statistic(
                x = x_data,
                values = y_data * weights,
                statistic = 'sum',
                bins = bins,
            )
            weights_sum, bin_edges, binnumber = scipy.stats.binned_statistic(
                x = x_data,
                values = weights,
                statistic = 'sum',
                bins = bins,
            )

            stat = weighted_sum / weights_sum

        else:

            assert weights is None, "weights only works with weighted_mean"

            # Usual statistic
            stat, bin_edges, binnumber = scipy.stats.binned_statistic(
                x = x_data,
                values = y_data,
                statistic = statistic,
                bins = bins,
            )

        # Calculate the percentiles
        def get_lower_percentile( data ):
            return np.percentile( data, lower_percentile )
        def get_upper_percentile( data ):
            return np.percentile( data, upper_percentile )
        low_p, bin_edges, binnumber = scipy.stats.binned_statistic(
            x = x_data,
            values = y_data,
            statistic = get_lower_percentile,
            bins = bins,
        )
        high_p, bin_edges, binnumber = scipy.stats.binned_statistic(
            x = x_data,
            values = y_data,
            statistic = get_upper_percentile,
            bins = bins,
        )

        # Get plotting axis
        if ax is None:
            fig = plt.figure( figsize=(10,9), facecolor='white' )
            ax = plt.gca()

        # X Values fo rplot
        x_values = bin_edges[:-1] + 0.5 * ( bin_edges[1] - bin_edges[0] )

        # Plot statistic
        ax.plot(
            x_values,
            stat,
            linewidth = linewidth,
            linestyle = linestyle,
            color = color,
            zorder = zorder,
            label = label,
        )

        # Plot interval
        if plot_interval:
            ax.fill_between(
                x_values,
                low_p,
                high_p,
                color = color,
                alpha = alpha,
            )

        # Add plot label
        if add_plot_label:
            if plot_label is None:
                plot_label = self.label
            if plot_label is not None:
                plt_label = ax.annotate(
                    s = plot_label,
                    **plot_label_kwargs
                 )

        if return_values:
            return stat, low_p, high_p, bin_edges

    ########################################################################

    def scatter(
        self,
        x_key, y_key,
        slices = None,
        n_subsample = None,
        ax = None,
        marker_size = 100,
        color = 'k',
        marker = '.',
        zorder = -100,
        x_range = None, y_range = None,
        x_label = None, y_label = None,
        add_x_label = True, add_y_label = True,
        plot_label = None,
        outline_plot_label = False,
        label_galaxy_cut = False,
        label_redshift = False,
        label_fontsize = 24,
        tick_param_args = None,
        out_dir = None,
        fix_invalid = True,
        line_slope = None,
        *args, **kwargs ):
        '''Make a 2D scatter plot of the data. Extra arguments are passed to get_selected_data.
        Args:
            x_key, y_key (str) : Data keys to plot.
            weight_key (str) : Data key for data to use as a weight. By None, no weight.
            slices (int or tuple of slices) : How to slices the data.
            ax (axis) : What axis to use. By None creates a figure and places the axis on it.
            x_range, y_range ( (float, float) ) : Histogram edges. If None, all data is enclosed. If list, set manually.
                If float, is +- x_range*length scale at that snapshot.
            n_bins (int) : Number of bins in the histogram.
            vmin, vmax (float) : Limits for the colorbar.
            plot_halos (bool) : Whether or not to plot merger tree halos on top of the histogram.
                Only makes sense for when dealing with positions.
            add_colorbar (bool) : If True, add a colorbar to colorbar_args
            colorbar_args (axis) : What axis to add the colorbar to. By None, is ax.
            x_label, ylabel (str) : Axes labels.
            add_x_label, add_y_label (bool) : Include axes labels?
            plot_label (str or dict) : What to label the plot with. By None, uses self.label.
                Can also pass a dict of full args.
            outline_plot_label (bool) : If True, add an outline around the plot label.
            label_galaxy_cut (bool) : If true, add a label that indicates how the galaxy was defined.
            label_redshift (bool) : If True, add a label indicating the redshift.
            label_fontsize (int) : Fontsize for the labels.
            tick_param_args (args) : Arguments to pass to ax.tick_params. By None, don't change inherent defaults.
            out_dir (str) : If given, where to save the file.
            fix_invalid (bool) : Fix invalid values.
            line_slope (float) : If given, draw a line with the given slope.
        '''

        if isinstance( slices, int ):
            sl = ( slice(None), slices )
        else:
            sl = slices

        # Get data
        x_data = self.data_object.get_selected_data( x_key, sl=sl, *args, **kwargs )
        y_data = self.data_object.get_selected_data( y_key, sl=sl, *args, **kwargs )

        # Fix NaNs
        if fix_invalid:
            x_mask = np.ma.fix_invalid( x_data ).mask
            y_mask = np.ma.fix_invalid( y_data ).mask
            mask = np.ma.mask_or( x_mask, y_mask )

            x_data = np.ma.masked_array( x_data, mask=mask ).compressed()
            y_data = np.ma.masked_array( y_data, mask=mask ).compressed()

        # Subsample
        if n_subsample is not None:
            sampled_inds = np.random.randint( 0, x_data.size, n_subsample )

            x_data = x_data[sampled_inds]
            y_data = y_data[sampled_inds]

        if x_range is None:
            x_range = [ x_data.min(), x_data.max() ]
        elif isinstance( x_range, float ):
            x_range = np.array( [ -x_range, x_range ])*self.data_object.ptracks.length_scale.iloc[slices]
        if y_range is None:
            y_range = [ y_data.min(), y_data.max() ]
        elif isinstance( y_range, float ):
            y_range = np.array( [ -y_range, y_range ])*self.data_object.ptracks.length_scale.iloc[slices]

        # Plot
        if ax is None:
            fig = plt.figure( figsize=(10,9), facecolor='white' )
            ax = plt.gca()

        s = ax.scatter( x_data, y_data, s=marker_size, color=color, marker=marker )

        # Change the z order
        s.set_zorder( zorder )

        # Halo Plot
        if line_slope is not None:
            line_x = np.array( [ x_data.min(), x_data.max() ] )
            line_y = line_slope*line_x
            ax.plot( line_x, line_y, linewidth=3, linestyle='dashed', )

        # Plot label
        if plot_label is None:
            plt_label = ax.annotate( s=self.label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, str ):
            plt_label = ax.annotate( s=plot_label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, dict ):
            plt_label = ax.annotate( **plot_label )
        elif plot_label is None:
            pass
        else:
            raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )
        if outline_plot_label:
            plt_label.set_path_effects([ path_effects.Stroke(linewidth=3, foreground='black'), path_effects.Normal() ])

        # Upper right label (info label)
        info_label = ''
        if label_galaxy_cut:
            info_label = r'$r_{ \rm cut } = ' + '{:.3g}'.format( self.data_object.galids.parameters['galaxy_cut'] ) + 'r_{ s}$'
        if label_redshift:
            info_label = r'$z=' + '{:.3f}'.format( self.data_object.ptracks.redshift.iloc[slices] ) + '$, '+ info_label
        if label_galaxy_cut or label_redshift:
            ax.annotate( s=info_label, xy=(1.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,
                ha='right' )

        # Add axis labels
        if add_x_label:
            if x_label is None:
                x_label = x_key
            ax.set_xlabel( x_label, fontsize=label_fontsize )
        if add_y_label:
            if y_label is None:
                y_label = y_key
            ax.set_ylabel( y_label, fontsize=label_fontsize )

        # Limits
        ax.set_xlim( x_range )
        ax.set_ylim( y_range )

        # Set tick parameters
        if tick_param_args is not None:
            ax.tick_params( **tick_param_args )

        # Save the file
        if out_dir is not None:
            save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[slices] )
            gen_plot.save_fig( out_dir, save_file, fig=fig, dpi=75 )

            plt.close()

    ########################################################################

    def plot_stacked_data(
        self,
        x_key,
        y_keys,
        colors,
        ax = None,
        *args, **kwargs
    ):
        
        if ax is None:
            plt.figure( figsize=(11, 5), facecolor='white' )
            ax = plt.gca()

        y_prev = np.zeros( shape=y_datas.values()[0].shape )

        y_datas = []
        for y_key in y_keys:
            y_data = self.data_object.get_selected_data(
                y_key,
                *args, **kwargs
            ).copy()

            y_datas.append( y_data )

        for i, y_key in y_keys:

            y_next = y_prev + y_datas[i]

            ax.fill_between(
                x_data,
                y_prev,
                y_next,
                color = classification_colors[key],
                alpha = p_constants.CLASSIFICATION_ALPHA,
            )

            # Make virtual artists to allow a legend to appear
            color_object = matplotlib.patches.Rectangle(
                (0, 0),
                1,
                1,
                fc = classification_colors[key],
                ec = classification_colors[key],
                alpha = p_constants.CLASSIFICATION_ALPHA,
            )
            color_objects.append( color_object )
            labels.append( p_constants.CLASSIFICATION_LABELS[key] )

        ax.annotate(
            s=self.label,
            xy=(0., 1.0225),
            xycoords='axes fraction',
            fontsize=22,
        )

        ax.legend(
            color_objects,
            labels,
            prop={'size': 14.5},
            ncol=5,
            loc=(0., -0.28),
            fontsize=20
        )

    ########################################################################

    def plot_time_dependent_data( self,
        ax = None,
        x_range = [ 0., np.log10(8.) ], y_range = None,
        y_scale = 'log',
        x_label = None, y_label = None,
        ):
        '''Make a plot like the top panel of Fig. 3 in Angles-Alcazar+17

        Args:
            ax (axis object) :
                What axis to put the plot on. By None, create a new one on a separate figure.

            x_range, y_range (list-like) :
                [ x_min, x_max ] or [ y_min, y_max ] for the displayed range.

            x_label, y_label (str) :
                Labels for axis. By None, redshift and f(M_star), respectively.

            plot_dividing_line (bool) :
                Whether or not to plot a line at the edge between stacked regions.
        '''

        if ax is None:
            fig = plt.figure( figsize=(11,5), facecolor='white' )
            ax = plt.gca()

        x_data = np.log10( 1. + self.data_object.get_data( 'redshift' ) )

        y_data = self.data_object.get_categories_galaxy_mass()

        for key in p_constants.CLASSIFICATION_LIST_A[::-1]:

            y_data = y_datas[key]

            ax.plot(
                x_data,
                y_data,
                linewidth = 3,
                color = p_constants.CLASSIFICATION_COLORS[key],
                label = p_constants.CLASSIFICATION_LABELS[key],
            )

        if x_range is not None:
            ax.set_xlim( x_range )

        if y_range is not None:
            ax.set_ylim( y_range )

        ax.set_yscale( y_scale )

        tick_redshifts = np.array( [ 0.25, 0.5, 1, 2, 3, 4, 5, 6, 7, ] )
        x_tick_values = np.log10( 1. + tick_redshifts )
        plt.xticks( x_tick_values, tick_redshifts )

        ax.set_xlabel( r'z', fontsize=22, )
        ax.set_ylabel( r'$M_{\star} (M_{\odot})$', fontsize=22, )

        ax.annotate( s=self.label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=22,  )

        ax.legend( prop={'size':14.5}, ncol=5, loc=(0.,-0.28), fontsize=20 )

    ########################################################################
    # Generic Plotting Methods
    ########################################################################

    def same_axis_plot(
        self,
        axis_plotting_method_str,
        variations,
        ax = None,
        figsize = (11, 5),
        out_dir = None,
        add_line_label = False,
        legend_args = { 'prop': {'size': 16.5}, 'loc': 'upper right', 'fontsize': 20 },
        *args, **kwargs
    ):

        if ax is None:
            fig = plt.figure( figsize=figsize, facecolor='white', )
            ax = plt.gca()

        all_plotting_kwargs = utilities.dict_from_defaults_and_variations( kwargs, variations )

        axis_plotting_method = getattr( self, axis_plotting_method_str )
        for key, plotting_kwargs in all_plotting_kwargs.items():

            plotting_kwargs['ax'] = ax

            if add_line_label:
                plotting_kwargs['line_label'] = key

            axis_plotting_method( *args, **plotting_kwargs )

        ax.legend( **legend_args )

        # Save the file
        if out_dir is not None:
            save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[kwargs['slices']] )
            gen_plot.save_fig( out_dir, save_file, fig=fig, dpi=75 )

            plt.close()

    ########################################################################

    def panel_plot( self,
        panel_plotting_method_str,
        defaults,
        variations,
        slices = None,
        n_rows = 2,
        n_columns = 2,
        plot_locations = [ (0,0), (0,1), (1,0), (1,1) ],
        figsize = (10,9),
        plot_label = None,
        outline_plot_label = False,
        label_galaxy_cut = False,
        label_redshift = True,
        label_fontsize = 24,
        subplot_label_args = { 'xy': (0.075, 0.88), 'xycoords': 'axes fraction', 'fontsize': 20, 'color': 'k',  },
        subplot_spacing_args = { 'hspace': 0.0001, 'wspace': 0.0001, },
        out_dir = None,
        ):
        '''
        Make a multi panel plot of the type of your choosing.

        Args:
            panel_plotting_method_str (str) : What type of plot to make.
            defaults (dict) : Default arguments to pass to panel_plotting_method.
            variations (dict of dicts) : Differences in plotting arguments per subplot.
            slices (slice) : What slices to select. By None, this doesn't pass any slices argument to panel_plotting_method
            plot_label (str or dict) : What to label the plot with. By None, uses self.label.
                Can also pass a dict of full args.
            outline_plot_label (bool) : If True, add an outline around the plot label.
            label_galaxy_cut (bool) : If true, add a label that indicates how the galaxy was defined.
            label_redshift (bool) : If True, add a label indicating the redshift.
            label_fontsize (int) : Fontsize for the labels.
            subplot_label_args (dict) : Label arguments to pass to each subplot for the label for the subplot.
                The actual label string itself corresponds to the keys in variations.
            subplot_spacing_args (dict) : How to space the subplots.
            out_dir (str) : If given, where to save the file.
        '''

        fig = plt.figure( figsize=figsize, facecolor='white', )
        ax = plt.gca()

        fig.subplots_adjust( **subplot_spacing_args )

        if slices is not None:
            defaults['slices'] = slices

        plotting_kwargs = utilities.dict_from_defaults_and_variations( defaults, variations )

        # Setup axes
        gs = gridspec.GridSpec(n_rows, n_columns)
        axs = []
        for plot_location in plot_locations:
            axs.append( plt.subplot( gs[plot_location] ) )

        # Setup arguments further
        for i, key in enumerate( plotting_kwargs.keys() ):
            ax_kwargs = plotting_kwargs[key]

            ax_kwargs['ax'] = axs[i]

            # Subplot label args
            this_subplot_label_args = subplot_label_args.copy()
            this_subplot_label_args['s'] = key
            ax_kwargs['plot_label'] = this_subplot_label_args

            if ax_kwargs['add_colorbar']:
                ax_kwargs['colorbar_args'] = { 'fig_or_ax': fig, 'ax_location': [0.9, 0.125, 0.03, 0.775 ],  }

            # Clean up interior axes
            ax_tick_parm_args = ax_kwargs['tick_param_args'].copy()
            plot_location = plot_locations[i]
            # Hide repetitive x labels
            if plot_location[0] != n_rows -1 :
                ax_kwargs['add_x_label'] = False
                ax_tick_parm_args['labelbottom'] = False
            # Hide repetitive y labels
            if plot_location[1] != 0:
                ax_kwargs['add_y_label'] = False
                ax_tick_parm_args['labelleft'] = False
            ax_kwargs['tick_param_args'] = ax_tick_parm_args

        # Actual panel plots
        panel_plotting_method = getattr( self, panel_plotting_method_str )
        for key in plotting_kwargs.keys():
            panel_plotting_method( **plotting_kwargs[key] )

        # Main axes labels
        # Plot label
        if plot_label is None:
            plt_label = axs[0].annotate( s=self.label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, str ):
            plt_label = axs[0].annotate( s=plot_label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, dict ):
            plt_label = axs[0].annotate( **plot_label )
        else:
            raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )
        if outline_plot_label:
            plt_label.set_path_effects([
                path_effects.Stroke(linewidth=3, foreground='white', background='white'),
                path_effects.Normal()
            ])

        # Upper right label (info label)
        info_label = ''
        if label_galaxy_cut:
            info_label = r'$r_{ \rm cut } = ' + '{:.3g}'.format( self.data_object.galids.parameters['galaxy_cut'] ) + 'r_{ s}$'
        if label_redshift:
            ind = defaults['slices']
            info_label = r'$z=' + '{:.3f}'.format( self.data_object.ptracks.redshift.iloc[ind] ) + '$'+ info_label
        if label_galaxy_cut or label_redshift:
            label_ax = plt.subplot( gs[0,n_columns-1,] )
            label_ax.annotate(
                s=info_label,
                xy=(1.,1.0225),
                xycoords='axes fraction',
                fontsize=label_fontsize,
                ha='right'
            )

        # Save the file
        if out_dir is not None:
            save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[slices] )
            gen_plot.save_fig( out_dir, save_file, fig=fig )

            plt.close()

    ########################################################################

    def make_multiple_plots( self,
        plotting_method_str,
        iter_args_key,
        iter_args,
        n_processors = 1,
        out_dir = None,
        make_out_dir_subdir = True,
        make_movie = False,
        clear_data = False,
        *args, **kwargs ):
        '''Make multiple plots of a selected type. *args and **kwargs are passed to plotting_method_str.

        Args:
            plotting_method_str (str) : What plotting method to use.
            iter_args_key (str) : The name of the argument to iterate over.
            iter_args (list) : List of argument values to change.
            n_processors (int) : Number of processors to use. Should only be used when saving the data.
            out_dir (str) : Where to save the data.
            make_movie (bool) : Make a movie out of the plots, if True.
            clear_data (bool) : If True, clear memory of the data after making the plots.
        '''

        plotting_method = getattr( self, plotting_method_str )

        if ( out_dir is not None ) and make_out_dir_subdir:
            out_dir = os.path.join( out_dir, self.label )

        def plotting_method_wrapper( process_args ):

            used_out_dir, used_args, used_kwargs = process_args

            plotting_method( out_dir=used_out_dir, *used_args, **used_kwargs )

            del used_out_dir, used_args, used_kwargs

            return

        all_process_args = []
        for iter_arg in iter_args:
            process_kwargs = dict( kwargs )
            process_kwargs[iter_args_key] = iter_arg
            all_process_args.append( ( out_dir, args, process_kwargs ) )

        if n_processors > 1:
            # For safety, make sure we've loaded the data already
            self.data_object.ptracks, self.data_object.galids, self.data_object.classifications

            mp_utils.parmap( plotting_method_wrapper, all_process_args, n_processors=n_processors, return_values=False )
        else:
            for i, iter_arg in enumerate( iter_args ):
                plotting_method_wrapper( all_process_args[i] )

        if make_movie:
            gen_plot.make_movie( out_dir, '{}_*.png'.format( self.label ), '{}.mp4'.format( self.label ), )

        if clear_data:
            del self.data_object.ptracks
            del self.data_object.galids
            del self.data_object.classifications

########################################################################
########################################################################

class PlotterSet( verdict.Dict ):
    '''Container for multiple plotters that is an enhanced dictionary.
    '''

    def __init__( self, data_object_cls, plotter_object_cls, defaults, variations ):
        '''
        Args:
            data_object_cls (object) : Class for the data object.
            plotter_object_cls (object) : Class for the plotter object.
            defaults (dict) : Set of None arguments for loading worldline data.
            variations (dict of dicts) : Labels and differences in arguments to be passed to Worldlines
        '''

        # Load the worldline sets
        storage = {}
        for key in variations.keys():

            kwargs = dict( defaults )
            for var_key in variations[key].keys():
                kwargs[var_key] = variations[key][var_key]

            storage[key] = { 'data_object': data_object_cls( **kwargs ), 'label': key }

        plotters_storage = utilities.SmartDict.from_class_and_args( plotter_object_cls, storage )

        super( PlotterSet, self ).__init__( plotters_storage )



#!/usr/bin/env python
'''Compilation of functions for interfacing with miscellanious data products.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import numpy as np
import os
import pandas as pd

########################################################################

def tidal_tensor_data_grudic(
    snum,
    ids = None,
    data_dir = '/work/03532/mgrudic/tidal_tensor',
):
    '''Load data Mike Grudic processed that contains Tidal Tensor, velocity
    dispersion, and items used for calculating the aforementioned quantities.

    Args:
        snum (int): Snapshot to retrieve the data for.

        ids (array-like): IDs to retrieve. Defaults to all.

        data_dir (str): Path to directory containing data.

    Returns:
        pandas.DataFrame
            DataFrame containing quantities. When given an ID not in the data
            returns NaN values for that ID.
    '''

    def invalid_data_result():
        '''Results when the data is invalid in some form.'''
        base_arr = np.full( len( ids ), np.nan )
        standin_data = {}
        data_keys = [
            'ID',
            'Txx',
            'Tyy',
            'Tzz',
            'Txy',
            'Tyz',
            'Tzx',
            'sigma_v',
            'r_search',
            'cond_num',
        ]
        for key in data_keys:
            standin_data[key] = copy.deepcopy( base_arr )
        standin_data['ID'] = ids
        df = pd.DataFrame( standin_data )
        df = df.set_index( 'ID' )

        return df

    # Load the data
    filename = 'tidal_tensor_{}.npy'.format( snum )
    file_path = os.path.join( data_dir, filename )
    try:
        full_arr = np.load( file_path )
    except FileNotFoundError:
        return invalid_data_result()
    
    # Convert to a pandas data frame to get the selected IDs out.
    data = {
        'ID': full_arr[:,0].astype( int ),
        'Txx': full_arr[:,1],
        'Tyy': full_arr[:,2],
        'Tzz': full_arr[:,3],
        'Txy': full_arr[:,4],
        'Tyz': full_arr[:,5],
        'Tzx': full_arr[:,6],
        'sigma_v': full_arr[:,7],
        'r_search': full_arr[:,8],
        'cond_num': full_arr[:,9],
    }
    df = pd.DataFrame( data, )
    df = df.set_index( 'ID' )

    # Select on IDs
    if ids is not None:
        try:
            df = df.loc[ids]
        except KeyError:
            return invalid_data_result()

    return df



#!/usr/bin/env python
'''Config file for galaxy diver.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import os

########################################################################

DIR = os.path.dirname( __file__ )
'''Where most of the source code is.'''

DATA_DIR = os.path.join( DIR, 'data' )
'''Where the data is store.'''



#!/usr/bin/env python
'''Tools for handling halo data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import glob
import numpy as np
import os
import pandas as pd
import unyt
import scipy.interpolate

import galaxy_dive.read_data.ahf as read_ahf
import galaxy_dive.read_data.rockstar as read_rockstar
import galaxy_dive.read_data.metafile as read_metafile
import galaxy_dive.utils.utilities as utilities
import galaxy_dive.analyze_data.generic_data as generic_data

########################################################################
########################################################################

class HaloData( generic_data.GenericData ):

    @utilities.store_parameters
    def __init__(
        self,
        data_dir,
        halo_finder = 'AHF',
        mt_kwargs = {},
        **kwargs
    ):
        '''Constructor for HaloData

        Args:
            data_dir (str) :
                Directory storing the data.

            halo_finder (str) :
                What code was the halo data generated with?
                Options: 'AHF', 'Rockstar'

            mt_kwargs (dict) :
                When loading merger tree halo files,
                additional arguments should be passed here.

            kwargs (dict) :
                When loading single-snapshot halo files, additional arguments
                should be passed here.
        '''

        if halo_finder == 'AHF':
            self.data_reader = read_ahf.AHFReader( data_dir )
        elif halo_finder == 'Rockstar':
            self.data_reader = read_rockstar.RockstarReader( data_dir )

        key_parser = HaloKeyParser()

        super( HaloData, self ).__init__( key_parser=key_parser )

    ########################################################################
    # Attributes
    ########################################################################

    def __getattr__( self, attr ):
        '''There are certain attributes held by self.data_reader that we
        would like to easily access here.
        '''

        data_reader_used_attrs = [
            'halos',
            'halos_add',
            'halos_snum',
            'mtree_halo_filepaths',
            'index', 
            'tag',
            'profiles',
            'profile_id_mapping',
        ]

        if attr in data_reader_used_attrs:
            return getattr( self.data_reader, attr )

    ########################################################################
    # Properties
    ########################################################################

    @property
    def mtree_halos( self ):
        '''Attribute for accessing merger tree data.
        '''

        if not hasattr( self.data_reader, 'mtree_halos' ):

            self.data_reader.get_mtree_halos( **self.mt_kwargs )

        return self.data_reader.mtree_halos

    ########################################################################

    @property
    def halos_path( self ):
        '''Location of the last loaded halo file.'''
        
        return self.data_reader.halos_path

    ########################################################################
    # Data Retrieval
    ########################################################################

    def get_data( self, data_key, snum, units=None ):
        '''Get halo data at a specific snapshot.

        Args:
            data_key (str) : What data to get.
            snum (int) : What snapshot to open.
            units (str) : If not None, convert data to these units.

        Returns:
            data (np.ndarray) : Requested data.
        '''

        # Check if the key is not unique to the halo data,
        # and if so, get the keys used for this particular data
        if data_key in self.data_reader.general_use_data_names.keys():
            used_data_key = self.data_reader.general_use_data_names[data_key]
        else:
            used_data_key = data_key

        # Load the necessary data
        self.get_halos( snum )

        # Special cases
        if used_data_key == 'ID':
            return self.data_reader.halos.index

        # Actual data
        data = self.data_reader.halos[used_data_key].values

        # Attach units if given
        if used_data_key in self.data_reader.units.keys():
            unit = self.data_reader.units[used_data_key]
            data *= unit

        # Convert to proper units, if necessary
        if units is not None:
            data.convert_to_units( units )
        
        return data

    ########################################################################

    def get_mt_data(
        self,
        data_key,
        mt_halo_id = 0,
        snums = None,
        a_power = None,
        return_values_only = True,
    ):
        '''Get halo data for a specific merger tree.

        Args:
            data_key (str) : What data to get.
            mt_halo_id (int) : What merger tree halo ID to select.
            snums (array-like) : If specified, get the values at these snapshots.
            a_power (float) : If given, multiply the result by the scale factor 1/(1 + redshift) to this power.
            return_values_only (bool) : If True, get rid of pandas data formatting

        Returns:
            mt_data (np.ndarray) : Requested data.
        '''

        mt_data = copy.copy( self.mtree_halos[mt_halo_id][data_key] )

        # For converting coordinates
        if a_power is not None:
            redshift = self.get_mt_data(
                'redshift',
                mt_halo_id,
                return_values_only = False
            )
            mt_data *= ( 1. + redshift )**-a_power

        if snums is not None:
            mt_data = mt_data.loc[snums]

        if return_values_only:
          mt_data = mt_data.values

        return mt_data

    ########################################################################

    def get_profile_data(
        self,
        data_key,
        snum,
        r,
        mt_halo_id = 0,
        a_power = None,
    ):
        '''Get halo data for a specific halo

        Args:
            data_key (str):
                What data to get.

            snum (int):
                What snapshot to open.

            r (float or arr of floats):
                What radii (in comoving kpc/h) to get data at.

            mt_halo_id (int):
                What merger tree halo ID to select.

            a_power (float):
                If given, multiply the result by the scale factor
                1/(1 + redshift) to this power.

        Returns:
            data (np.ndarray) : Requested data.
        '''

        # Cast as an array for easier manipulation
        r = np.array( r )

        # Set up the base result
        data = np.full( r.size, np.nan )

        # Load the necessary data
        self.data_reader.get_profiles( snum )

        # Replace the merger tree halo ID with the relevant redshift halo ID
        halo_id = self.get_mt_data( 'ID', snums=[snum] )[0]

        # Get the profile for the right halo
        start_ind = self.profile_id_mapping[halo_id]
        end_ind = self.profile_id_mapping[halo_id+1]
        halo_profile = self.profiles.loc[start_ind:end_ind-1]

        # Find the valid regime
        halo_r_min = np.abs( np.min( halo_profile['r'] ) )
        halo_r_max = np.max( halo_profile['r'] )
        r_not_too_small = r > halo_r_min
        r_not_too_large = r < halo_r_max
        valid = r_not_too_small & r_not_too_large

        # Get the results out
        interp_fn = scipy.interpolate.interp1d(
            halo_profile['r'].values,
            halo_profile[data_key].values,
        )
        data[valid] = interp_fn( r[valid] )

        # For converting coordinates
        if a_power is not None:
            z = self.get_mt_data( 'redshift', mt_halo_id, snums=snum )[0]
            data *= ( 1. + z )**-a_power

        return data

    ########################################################################

    def get_enclosed_mass(
        self,
        positions,
        snum,
        hubble_param,
        mt_halo_id = 0,
    ):
        '''Get the mass enclosed at the positions given.
        Particles must be inside a halo, and the enclosed mass is only the mass
        of that halo.

        Args:
            positions (array-like, (n_positions,3)):
                Positions in physical kpc you would like to
                find the mass interior to. Positions should be relative to
                the simulation box coordinates.

            snum (int):
                Snapshot at which to get the enclosed mass.

            hubble_param (float):
                The hubble parameter, H0/100, needed for converting to and from
                AHF's units.
            
            mt_halo_id (int):
                Merger tree halo ID for the halo you want to use for finding
                the enclosed mass.

        Returns:
            array-like, (n_positions,):
                The mass of Merger Traced halo `mt_halo_id` interior to
                the positions given. A value of np.nan indicates that the
                position is either outside the halo or too far in (and the
                results aren't converged).
        '''

        # Find the radii
        origin = np.array([
            self.get_mt_data( 'Xc', a_power=1., snums=[snum], mt_halo_id=mt_halo_id )[0],
            self.get_mt_data( 'Yc', a_power=1., snums=[snum], mt_halo_id=mt_halo_id )[0],
            self.get_mt_data( 'Zc', a_power=1., snums=[snum], mt_halo_id=mt_halo_id )[0],
        ]) / hubble_param
        centered_positions = positions - origin
        r = np.sqrt( (centered_positions**2. ).sum( axis=1 ) )

        # Convert to comoving kpc/h
        z = self.get_mt_data( 'redshift', mt_halo_id, snums=[snum] )[0]
        r *= hubble_param * ( 1. + z )

        # Get the profile data out
        enclosed_mass = self.get_profile_data(
            'M_in_r',
            snum,
            r,
            mt_halo_id = mt_halo_id,
        )

        return enclosed_mass / hubble_param

    ########################################################################

    def get_halos( self, snum, **kwargs ):
        '''Get halo data.

        Args:
            snum (int) : What snapshot to get the halo data for.

            **kwargs :
                Additional arguments for loading the data.
                In the event of conflict with those provided when constructing
                the instance, arguments provided here will be preferred.
        '''

        # Get the used arguments
        used_kwargs = utilities.merge_two_dicts(
            kwargs,
            self.kwargs,
        )

        # Try to load additional postprocessed data
        self.data_reader.get_halos( snum, **used_kwargs )
        if hasattr( self.data_reader, 'get_halos_add' ):
            try:
                self.data_reader.get_halos_add( snum, **used_kwargs )
            except NameError:
                pass

    ########################################################################

    def get_mtree_halos( self, **kwargs ):
        '''Get halo data.

        Args:
            snum (int) : What snapshot to get the halo data for.

            **kwargs :
                Additional arguments for loading the data.
                In the event of conflict with those provided when constructing
                the instance, arguments provided here will be preferred.
        '''

        # Get the used arguments
        used_kwargs = utilities.merge_two_dicts(
            kwargs,
            self.mt_kwargs,
        )

        # Try to load additional postprocessed data
        self.data_reader.get_mtree_halos( **used_kwargs )

    ########################################################################

    def get_n_halos( self, snum ):
        '''Get the number of halos (non-tracked) at a given redshift.'''

        self.get_halos( snum )

        return self.data_reader.halos.index.size

    ########################################################################

    def get_selected_data( self, *args, **kwargs ):

        return super( HaloData, self ).get_selected_data(
            mask_multidim_data = False,
            *args, **kwargs
        )

########################################################################
########################################################################

class HaloKeyParser( generic_data.DataKeyParser ):

    def get_radius_key( self, multiplier, length_scale ):
        '''Get a key for Halo data, based on a length scale and a multiple of it.

        Args:
            multiplier (float) :
                multiplier*length_scale defines the radius around the center of the halo(s).

            length_scale (str) :
                multiplier*length_scale defines the radius around the center of the halo(s).

        Returns:
            radius_key (str) :
                Combination of length_scale and multiplier.
        '''

        if np.isclose( multiplier, 1.0 ):
            radius_key = length_scale
        else:
            radius_key = '{}{}'.format( multiplier, length_scale )

        return radius_key

    ########################################################################

    def get_enclosed_mass_key( self, ptype, multiplier, length_scale ):
        '''Get a key for Halo data, corresponding to a data column that records an enclosed mass.

        Args:
            ptype (str) :
                The particle type for the enclosed mass.

            multiplier (float) :
                multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

            length_scale (str) :
                multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

        Returns:
            enclosed_mass_key (str)
        '''

        return 'M{}({})'.format( ptype, self.get_radius_key( multiplier, length_scale ) )

    ########################################################################

    def get_average_quantity_key( self, data_key, ptype, multiplier, length_scale ):
        '''Get a key for Halo data, corresponding to a data column that records the average quantity inside a galaxy.

        Args:
            data_key (str) :
                What the enclosed quantity is.

            ptype (str) :
                The particle type for the enclosed mass.

            multiplier (float) :
                multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

            length_scale (str) :
                multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

        Returns:
            average_quantity_key (str)
        '''

        return '{}{}({})'.format( data_key, ptype, self.get_radius_key( multiplier, length_scale ) )


    ########################################################################

    def get_velocity_at_radius_key( self, velocity_key, multiplier, length_scale ):
        '''Get a key for Halo data, corresponding to a data column that records the velocity at at a specified radius

            velocity_key (str) :
                What velocity to get.

            multiplier (float) :
                multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

            length_scale (str) :
                multiplier*length_scale defines the radius around the center of the halo within which to get the mass.

        Returns:
            velocity_at_radius_key (str)
        '''

        return '{}({})'.format( velocity_key, self.get_radius_key( multiplier, length_scale ) )




#!/usr/bin/env python
'''Tools for reading simulation metafiles.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import glob
import numpy as np
import os
import pandas as pd

########################################################################
########################################################################

class MetafileReader( object ):
  '''Read simulation metafiles, e.g. snapshot_times.txt
  '''

  def __init__( self, sdir ):
    '''Initializes.

    Args:
      sdir (str): Simulation directory to load the metafiles from.
    '''

    self.sdir = sdir

  ########################################################################

  def get_snapshot_times( self ):
    '''Load the snapshot_times.txt files that are in the simulation directories.

    Modifies:
      self.snapshot_times (pd.DataFrame): A dataframe containing information about the snapshot times.
    '''

    # FIRE-2 snapshot times
    # Assumes the file is named snapshot_times.txt
    try:
      filepath = os.path.join( self.sdir, 'snapshot_times.txt' )
      
      # Column names
      names = [ 'snum', 'scale-factor', 'redshift', 'time[Gyr]', 'time_width[Myr]' ]

      self.snapshot_times = pd.read_csv( filepath, delim_whitespace=True, skiprows=3, index_col=0, names=names ) 

    # FIRE-1 snapshot times (these are simpler)
    # Assumes the file is named output_times.txt
    except:

      filepath = os.path.join( self.sdir, 'output_times.txt' )

      # Column names
      names = [ 'scale-factor', ]

      # Load the data
      self.snapshot_times = pd.read_csv( filepath, names=names )

      # Rename the index
      self.snapshot_times.index.name = 'snum'

      # Get the redshift
      self.snapshot_times['redshift'] = 1./self.snapshot_times['scale-factor'] - 1.

  ########################################################################

  def get_used_parameters( self ):
    '''Load parameters used to run the simulation from, e.g. the gizmo_parameters.txt-usedvalues file.

    Modifies:
      self.used_parameters( pd.DataFrame): A dataframe containing the parameters.
    '''

    potential_filepaths = glob.glob( '{}/*usedvalues'.format( self.sdir ) )

    assert len( potential_filepaths ) < 2, 'Multiple options to choose the parameter file from.'
    assert len( potential_filepaths ) != 0, 'No used parameter file found (e.g. gizmo_parameters.txt-usedvalues).'

    parameter_filepath = potential_filepaths[0]

    self.used_parameters = {}
    with open(parameter_filepath,'r') as input_file:

      # Loop through the lines
      for i, line in enumerate( input_file ):

        # Split the line up
        split_line = line.split()
        
        # Check for lines that don't fit the expectations
        assert len( split_line ) == 2, 'Unexpected format in Line {}'.format( i )

        self.used_parameters[split_line[0]] = split_line[1]





#!/usr/bin/env python
'''Subclass for analyzing particle data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import numpy as np

import galaxy_dive.analyze_data.simulation_data as simulation_data
import galaxy_dive.read_data.snapshot as read_snapshot
import galaxy_dive.utils.constants as constants
import galaxy_dive.utils.utilities as utilities

########################################################################

class ParticleData( simulation_data.SnapshotData ):
  '''Subclass for particle data.
  '''

  @utilities.store_parameters
  def __init__( self, sdir=None, snum=None, ptype=None, load_additional_ids=False, **kwargs ):
    '''Initialize.

    Args:
      sdir (str) : Directory the snapshot is stored in
      snum (int) : Snapshot number to open
      ptype (int) : Particle type to load.
      load_additional_ids (bool, optional) : Whether or not to load child_ids, etc, if they exist.
    '''

    super( ParticleData, self ).__init__( data_dir=sdir, snum=snum, **kwargs )

    self.retrieve_data()

  ########################################################################

  def retrieve_data( self ):

    # Assume we convert from cosmological units
    P = read_snapshot.readsnap( self.sdir, self.snum, self.ptype, load_additional_ids=self.load_additional_ids,
                                cosmological=True, )

    # Parse the keys and put in a more general format
    # All units should be the standard GIZMO output units
    self.data = {}
    self.data_attrs = {}
    for key in P.keys():

      # Get the attributes
      attrs_keys = ['redshift', 'omega_lambda', 'flag_metals', 'flag_cooling', 'omega_matter', 'flag_feedbacktp',
                    'time', 'boxsize', 'hubble', 'flag_sfr', 'flag_stellarage', 'k']
      if key in attrs_keys:
        self.data_attrs[key] = P[key]

      # Get the data
      # Gas Density
      elif key == 'rho':
        self.data['Den'] = P[key]
      # Gas Neutral Hydrogen fraction
      elif key == 'nh':
        self.data['nHI'] = P[key]
      # Should be the mean free electron number per proton
      elif key == 'ne':
        self.data['ne'] = P[key]
      # Star Formation Rate
      elif key == 'sfr':
        self.data['SFR'] = P[key]
      # Position
      elif key == 'p':
        self.data['P'] = P[key].transpose()
      # Velocity
      elif key == 'v':
        self.data['V'] = P[key].transpose()
      # Metal mass fraction
      elif key == 'z':
        self.data['Z'] = P[key][:,0] # Total metallicity (everything not H or He)
        self.data['Z_Species'] = P[key][:,1:] # Details per species, [He, C, N, O, Ne, Mg, Si, S, Ca, Fe], in order
      # Particle IDs
      elif key == 'id':
        self.data['ID'] = P[key]
      elif key == 'child_id':
        self.data['ChildID'] = P[key]
      elif key == 'id_gen':
        self.data['IDGen'] = P[key]
      # Particle Masses
      elif key == 'm':
        self.data['M'] = P[key]
      # Internal energy
      elif key == 'u':
        self.data['U'] = P[key]
      # Smoothing lengths
      elif key == 'h':
        self.data['h'] = P[key]
      elif key == 'age':
        self.data['Age'] = P[key]
      elif key == 'potential':
        self.data['Potential'] = P[key]
      else:
        raise Exception('NULL key, key={}'.format(key))

  ########################################################################

  def calc_temp( self, gamma=5./3. ):
    '''Calculate the temperature from the internal energy. '''

    #raise Exception( TODO: Test this )

    mu = self.calc_mu()
    u_cgs = self.data['U']*1.e10
    self.data['T'] = constants.PROTON_MASS*mu*(gamma - 1)*u_cgs/constants.K_B

  ########################################################################

  def calc_mu( self ):
    '''we track this with metal species now, could do better...
    '''

    XH=0.76

    yhelium = ( 1. - XH ) / ( 4. * XH )

    return ( 1. + 4. * yhelium ) / ( 1. + yhelium + self.data['ne'] )

  ########################################################################

  def calc_pressure( self ):
    '''Calculate the pressure in units of K/cm^3.'''

    self.data['Pressure'] = self.get_data( 'NumDen' )*self.get_data( 'T' )

  ########################################################################

  def calc_classifications( self ):
    '''Get the classification for each particle, using data from the Angles-Alcazar+16 pipeline.
    Uses classes from the tracked_particle_data_handling.py module.

    Parameters (include in data_p)
    'tracked_p_data_dir' : Directory containing the tracked-particle data.
    'tracked_p_file_tag' : Identifying tag for the tracked-particle data.
    '''

    sim_name = self.data_p['sdir'].split( '/' )[-1]
    full_data_dir = os.path.join( self.data_p['tracked_p_data_dir'], sim_name )

    # Load the actual tracked particle data
    tracked_p_data = tracked_particle_data_handling.TrackedParticleDataHandler(full_data_dir,
                                                                               self.data_p['tracked_p_file_tag'])

    # Get the classifications
    self.data['Cl'] = tracked_p_data.classify_dataset(self.data['ID'], self.data_attrs['redshift'])

  ########################################################################

  def calc_metal_mass( self ):
      
    self.data['metal_mass'] = self.get_data( 'M' ) * self.get_data( 'Z' )

  ########################################################################

  def calc_enriched_metal_mass( self ):
    '''Calculate the metal mass that comes from enrichment, not counting mass
    that's at the metallicity floor. Assumes that the there will always be at
    least one particle in the simulation that's at the metallicity floor.'''

    enrichment_fraction = self.get_data( 'Z' ) - self.get_data( 'Z' ).min()

    self.data['enriched_metal_mass'] = self.get_data( 'M' ) * enrichment_fraction

  ########################################################################

  def find_duplicate_ids( self ):
    '''Find all the IDs in the particle data set that have duplicates.

    Returns:
      duplicate_ids ( np.ndarray ) : Array of IDs that have duplicates.
    '''

    unique_ids, id_counts = np.unique( self.get_data( 'ID' ), return_counts=True )

    duplicate_inds = np.where( id_counts > 1 )[0]

    return unique_ids[duplicate_inds]
































#!/usr/bin/env python
'''Class for analyzing simulation data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

# Base python imports
import copy
from functools import wraps
import h5py
import numpy as np
import numpy.testing as npt
import warnings
import galaxy_dive.utils.constants as constants
import galaxy_dive.utils.utilities as utilities

########################################################################

# For default values
default = object()

########################################################################
########################################################################

class GenericData( object ):
  '''Very generic data class, with getting and masking functionality.

    Args:
      key_parser (object) : KeyParser instance to use to interpret data keys.
      data_masker (object) : DataMasker instance to use to filter and mask data.
      verbose (bool) : Print out additional information.
      z_sun (float) : Used mass fraction for solar metallicity.
    '''

  @utilities.store_parameters
  def __init__( self,
    key_parser = None,
    data_masker = None,
    verbose = False,
    z_sun = constants.Z_MASSFRAC_SUN,
    **kwargs ):

    # For storing and creating masks to pass the data
    if data_masker is None:
      self.data_masker = DataMasker( self )

    # Setup a data key parser
    if key_parser is None:
      self.key_parser = DataKeyParser()

  ########################################################################
  # Properties
  ########################################################################

  @property
  def length_scale( self ):
    '''Property for fiducial length scale. By default, is 1.
    However, a more advanced subclass might set this differently, or this might
    change in the future.
    '''

    # TODO: Address this.
    raise Exception( "Current thinking: this should not be called." )

    return 1.

  ########################################################################

  @property
  def velocity_scale( self ):
    '''Property for fiducial velocity scale. By default is 1.
    However, a more advanced subclass might set this differently, or this might
    change in the future.
    '''

    # TODO: Address this.
    raise Exception( "Current thinking: this should not be called." )

    return 1.

  ########################################################################

  @property
  def metallicity_scale( self ):
    '''Property for fiducial metallicity scale. By default is z_sun
    However, a more advanced subclass might set this differently, or this might
    change in the future.
    '''

    return self.z_sun

  ########################################################################

  @property
  def base_data_shape( self ):
    '''Property for simulation redshift.'''

    if not hasattr( self, '_base_data_shape' ):
      self._base_data_shape = self.data.values()[0].shape

    return self._base_data_shape

  @base_data_shape.setter
  def base_data_shape( self, value ):
    '''Setting function for simulation redshift property.'''

    # If we try to set it, make sure that if it already exists we don't change it.
    if hasattr( self, '_base_data_shape' ):
      assert self._base_data_shape == value

    else:
      self._base_data_shape = value

  ########################################################################
  # Data Retrieval
  ########################################################################

  def get_data( self, data_key, sl=None, data_storage=default ):
    '''Get the data from within the class. Only for getting the data. No post-processing or changing the data
    (putting it in particular units, etc.) The idea is to calculate necessary quantities as the need arises,
    hence a whole function for getting the data.

    Args:
      data_key (str) : Key in the data dictionary for the key we want to get
      sl (slice) : Slice of the data, if requested.

    Returns:
      data (np.ndarray) : Requested data.
    '''

    if data_storage is default:
      data_storage = self.data

    # Loop through, handling issues
    n_tries = 10
    for i in range( n_tries ):
      try:

        # Arbitrary functions of the data
        if data_key == 'Function':

          raise Exception( "TODO: Test this" )

          # Use the keys to get the data
          function_data_keys = self.kwargs['function_args']['function_data_keys']
          input_data = [ self.get_data(function_data_key) for function_data_key in function_data_keys ]

          # Apply the function
          data = self.kwargs['function_args']['function']( input_data )

        # Other
        else:
          data = self.data[data_key]

      # Calculate missing data
      except KeyError as e:
        self.handle_data_key_error( data_key )
        continue

      break

    if 'data' not in locals().keys():
      raise KeyError( "After {} tries, unable to find or create data_key, {}".format( i+1, data_key ) )

    if sl is not None:
      return data[sl]

    return data

  ########################################################################

  def get_processed_data( self, data_key, data_method=default, *args, **kwargs ):
    '''Get post-processed data. (Accounting for fractions, log-space, etc.).

    Args:
      data_key (str) : What data to get.
      data_method (str) : What method to use for getting the data itself. Defaults to using self.get_data
      *args, **kwargs : Passed to get_data()

    Returns:
      processed_data (np.ndarray) : Requested data, including formatting.
    '''

    # Account for fractional data keys
    data_key, fraction_flag = self.key_parser.is_fraction_key( data_key )

    # Account for logarithmic data
    data_key, log_flag = self.key_parser.is_log_key( data_key )

    # Choose what method we're using for getting data.
    if data_method is default:
      get_data_method = self.get_data
    else:
      get_data_method = getattr( self, data_method )

    # Get the data and make a copy to avoid altering
    data = copy.deepcopy( get_data_method( data_key, *args, **kwargs ) )

    # Actually calculate the fractional data
    if fraction_flag:

      # Put distances in units of the virial radius
      if self.key_parser.is_position_key( data_key ):

        data /= self.length_scale

      # Put velocities in units of the circular velocity
      elif self.key_parser.is_velocity_key( data_key ):
        data /= self.velocity_scale

      # Put the metallicity in solar units
      elif data_key == 'Z':
        data /= self.metallicity_scale

      else:
        raise Exception('Fraction type not recognized')

    # Make appropriate units into log
    if log_flag:
      data =  np.log10( data )

    return data

  ########################################################################

  def get_selected_data( self, *args, **kwargs ):
    '''Wrapper for getting masked data.'''

    return self.data_masker.get_selected_data( *args, **kwargs )

  def mask_data( self, *args, **kwargs ):
    '''Wrapper for masking data.'''

    return self.data_masker.mask_data( *args, **kwargs )

  ########################################################################

  def shift( self, data, data_key ):
    '''Shift or multiply the data by some amount. Note that this is applied after logarithms are applied.

    data : data to be shifted
    data_key : Data key for the parameters to be shifted
    Parameters are a subdictionary of self.kwargs
    'data_key' : What data key the data is shifted for
    '''

    raise Exception( "TODO: Test this" )

    shift_p = self.kwargs['shift']

    # Exit early if not the right data key
    if shift_p['data_key'] != data_key:
      return 0

    # Shift by the mass metallicity relation.
    if 'MZR' in shift_p:

      # Gas-phase ISM MZR fit from Ma+2015.
      # Set to be no shift for gas with the mean metallicity of LLS at z=0 (This has a value of -0.45, as I last calculated)
      log_shift = 0.93*(np.exp(-0.43*self.redshift) - 1.) - 0.45

      data -= log_shift

  ########################################################################

  def handle_data_key_error( self, data_key ):
    '''Method for attempting to generate data on the fly.

    Args:
      data_key (str) : Type of data to attempt to generate data for.
    '''

    raise Exception( "This method should be replaced in the subclass!" )

  ########################################################################
  # Meta Methods (for doing generic things with class methods)
  ########################################################################

  def iterate_over_method( self, method_str, iter_arg, iter_values, method_args ):
    '''Iterate over a specified method, and get the results out.

    Args:
      method_str (str) :
        Which method to use.

      iter_arg (str) :
        Which argument of the method to iterate over.

      iter_values (list of values) :
        Which values to change.

      method_args (dict) :
        Default args to pass to the method

    Returns:
      results (list) : [ method( **used_args ) for used_args in all_variations_of_used_args ]
    '''

    method = getattr( self, method_str )

    results = []
    for iter_value in iter_values:

      method_args[iter_arg] = iter_value

      result = method( **method_args )

      results.append( result )

    return results

########################################################################
########################################################################

class DataKeyParser( object ):
  '''Class for parsing data_keys provided to SimulationData.'''

  ########################################################################

  def is_position_key( self, data_key ):
    '''Checks if the data key deals with position primarily.

    Args:
      data_key (str) : Data key to check.

    Returns:
      is_position_key (bool) : True if deals with position.
    '''

    return ( data_key[0] == 'R' ) | ( data_key == 'P' )

  ########################################################################

  def is_velocity_key( self, data_key ):
    '''Checks if the data key deals with velocity primarily.

    Args:
      data_key (str) : Data key to check.

    Returns:
      is_velocity_key (bool) : True if deals with velocity.
    '''

    return ( data_key[0] == 'V' )


  ########################################################################

  def is_fraction_key( self, data_key ):
    '''Check if the data should be some fraction of its relevant scale.

    Args:
      data_key (str) : Data key to check.

    Returns:
      is_fraction_key (bool) : True if the data should be scaled as a fraction of the relevant scale.
    '''

    fraction_flag = False
    if data_key[-1] == 'f':
      data_key = data_key[:-1]
      fraction_flag = True

    return data_key, fraction_flag

  ########################################################################

  def is_log_key( self, data_key ):
    '''Check if the data key indicates the data should be put in log scale.

    Args:
      data_key (str) : Data key to check.

    Returns:
      is_log_key (bool) : True if the data should be taken log10 of
    '''

    log_flag = False
    if data_key[0:3] == 'log':
      data_key = data_key[3:]
      log_flag = True

    return data_key, log_flag

########################################################################
########################################################################

class DataMasker( object ):

  def __init__( self, data_object ):
    '''Class for masking data.

    Args:
      data_object (GenericData object) : Used for getting data to find mask ranges.
    '''

    self.data_object = data_object

    self.masks = []
    self.optional_masks = {}

  ########################################################################

  def mask_data( self,
    data_key,
    data_min = default,
    data_max = default,
    data_value = default,
    custom_mask = default,
    return_or_store = 'store',
    optional_mask = False,
    mask_name = default,
    *args, **kwargs ):
    '''Get only the particle data within a certain range. Note that it retrieves the processed data.

    Args:
      data_key (str) :
        Data key to base the mask off of.

      data_min (float) :
        Everything below data_min will be masked.

      data_max (float) :
        Everything above data_max will be masked.

      data_value (float) :
        Everything except for data_value will be masked.

      custom_mask (bool) :
        If provided, take in a custom mask instead, using data_key as the label for the mask.

      return_or_store (str) :
        Whether to store the mask as part of the masks dictionary, or to return it.

      optional_mask (bool) :
        If True, store in the dictionary self.optional_masks instead.

      mask_name (str) :
        What name to associate with this mask? Currently only relevant if optional_mask is True.
        By default uses the data_key as the name.

      *args, **kwargs :
        Passed to self.data_object.get_processed_data()

    Returns:
      data_mask (np.array of bools) :
        If requested, the mask for data in that range.

    Modifies:
      self.masks (list of dicts) :
        Appends a dictionary describing the mask.
    '''

    # Process what type of mask to get
    mask_outside = ( data_min is not default ) and ( data_max is not default )
    mask_discrete = data_value is not default
    mask_custom = custom_mask is not default
    assert ( mask_outside + mask_discrete + mask_custom )==1, "Bad combination of masks!"

    if not mask_custom:
      data = self.data_object.get_processed_data( data_key, *args, **kwargs )

    # Get the mask
    if mask_outside:
      data_ma = np.ma.masked_outside( data, data_min, data_max )
      mask = data_ma.mask
    elif mask_discrete:
      mask = np.invert( data == data_value )
    elif mask_custom:
      mask = custom_mask
    else:
      raise NameError( "Unspecified combination of data masking." )

    # Handle the case where an entire array is masked or none of it is masked
    # (Make into an array for easier combination with other masks)
    if mask.size == 1:
      mask = mask*np.ones( shape=data.shape, dtype=bool )

    if return_or_store == 'store':

      if mask_outside:
        mask_dict = {'data_key': data_key, 'data_min': data_min, 'data_max': data_max, 'mask': mask}
      elif mask_discrete:
        mask_dict = {'data_key': data_key, 'data_value' : data_value, 'mask': mask}
      elif mask_custom:
        mask_dict = {'data_key': data_key, 'custom_mask' : True, 'mask': mask}

      if optional_mask:
        if mask_name is default:
          mask_name = data_key

        assert mask_name not in self.optional_masks.keys(), "A mask with that name already exists!"

        self.optional_masks[mask_name] = mask_dict

      else:
        self.masks.append( mask_dict )

    elif return_or_store == 'return':
      return mask

    else:
      raise Exception('NULL return_or_store')

  ########################################################################

  def get_total_mask( self, optional_masks=None ):
    '''Get the result of combining all masks in the data.

    Args:
      optional_masks (list-like) : List of names of optional masks to use (must be found in self.optional_masks).

    Returns:
      total_mask (np.array of bools) : Result of all masks.
    '''

    # Compile masks
    all_masks = []
    for mask_dict in self.masks:
      all_masks.append( mask_dict['mask'] )

    # Get any requested optional masks
    if optional_masks is not None:
      for optional_mask in optional_masks:
        all_masks.append( self.optional_masks[optional_mask]['mask'] )

    # Combine masks
    return np.any( all_masks, axis=0, keepdims=True )[0]

  ########################################################################

  def get_selected_data( self,
    data_key,
    mask = 'total',
    optional_masks = None,
    sl = None,
    apply_slice_to_mask = True,
    fix_invalid = False,
    compress = True,
    mask_multidim_data = True,
    *args, **kwargs ):
    '''Get all the data that doesn't have some sort of mask applied to it. Use the processed data.

    Args:
      data_key (str) : Data key to get the data for.
      mask (str or np.array of bools) : Mask to apply. If none, use the total mask.
      optional_masks (list-like) : List of names of optional masks to use (must be found in self.optional_masks).
      sl (slice) : Slice to apply to the data
      apply_slice_to_mask (bool) : Whether or not to apply the same slice you applied to the data to the mask.
      fix_invalid (bool) : Whether or not to also mask invalid data.
      compress (bool) : Whether or not to return compressed data.
      mask_multidim_data (bool) : Whether or not to change the mask to fit multidimensional data.
      *args, **kwargs : Passed to get_proceesed_data.

    Returns:
      data_ma (np.array) : Compressed masked data. Because it's compressed it may not have the same shape as the
        original data.
    '''

    data = self.data_object.get_processed_data( data_key, sl=sl, *args, **kwargs )

    # Get the appropriate mask
    if isinstance( mask, np.ndarray ):
      used_mask = mask
    elif isinstance( mask, bool ) or isinstance( mask, np.bool_ ):
      if not mask:
        if fix_invalid:
          return np.ma.fix_invalid( data ).compressed()
        else:
          return data

      raise Exception( "All data is masked." )
    elif mask == 'total':
      used_mask = self.get_total_mask( optional_masks=optional_masks )
    else:
      raise KeyError( "Unrecognized type of mask, {}".format( mask ) )

    if ( sl is not None ) and apply_slice_to_mask:
      used_mask = used_mask[sl]

    if fix_invalid:
      array_to_ma_array_fn = np.ma.fix_invalid
    else:
      array_to_ma_array_fn = np.ma.array

    # Test for if the data fits the mask, or if it's multi-dimensional
    if mask_multidim_data:
      if len( data.shape ) > len( self.data_object.base_data_shape ):
        data_ma = [ array_to_ma_array_fn( data_part, mask=used_mask ) for data_part in data ]
        data_ma = [ data_ma_part.compressed() for data_ma_part in data_ma ]
        data_ma = np.array( data_ma )

      else:
        data_ma = array_to_ma_array_fn( data, mask=used_mask )

        if compress:
          data_ma = data_ma.compressed()

    else:
      data_ma = array_to_ma_array_fn( data, mask=used_mask )

      if compress:
        data_ma = data_ma.compressed()

    return data_ma

  ########################################################################

  def clear_masks( self, clear_optional_masks=False ):
    '''Reset the masks in total to nothing.

    Modifies:
      self.masks (lists) : Sets to empty
    '''

    self.masks = []

    if clear_optional_masks:
      self.optional_masks = {}



#!/usr/bin/env python
'''Testing for trends.cgm.py

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import glob
from mock import call, patch
import numpy as np
import numpy.testing as npt
import os
import pdb
import pytest
import unittest

import galaxy_dive.trends.data_products as data_products

########################################################################
########################################################################

# Decorator for skipping tests unique to stampede
stampede = pytest.mark.skipif(
    not pytest.config.getoption("--runstampede"),
    reason="need --runstampede option to run"
)

@stampede
class TestDataProducts( unittest.TestCase ):

    def test_tidal_tensor_data_grudic( self ):
        '''Just test that we're reading the data correctly.
        '''

        actual = data_products.tidal_tensor_data_grudic(
            100,
            ids = np.array([ 48850256, 9550185, 1111111111111 ]),
        )
        # From manually looking at the data
        expected = np.array([
            [ 1.22668942e+03,  7.75438811e+01,  6.02076014e+02, -3.91015235e+01, -1.17870247e+02,  7.95706269e+01,  8.74140108e+01, 1.31098915e-01,  6.59692747e+03 ],
            [ 1.00920133e+01,  4.79382207e+01,  2.02634658e+01, 6.64207692e+00,  2.51895674e+01, -1.50385106e+01,  3.50740113e+01, 1.26965516e+00,  5.80723533e+01 ],
            [ np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ],
        ])

        npt.assert_allclose( expected, actual.values )
            
    ########################################################################

    def test_tidal_tensor_data_grudic_no_matching_ids( self ):
        '''Test what happens when no IDs match.
        '''

        actual = data_products.tidal_tensor_data_grudic(
            100,
            ids = np.array([ 1111111111111, 1111111111112, 1111111111113 ]),
        )
        # From manually looking at the data
        expected = np.array([
            [ np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ],
            [ np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ],
            [ np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ],
        ])

        npt.assert_allclose( expected, actual.values )
            
    ########################################################################

    def test_tidal_tensor_data_grudic_no_data( self ):
        '''Test what happens when a file doesn't exist
        '''

        actual = data_products.tidal_tensor_data_grudic(
            10000,
            ids = np.array([ 48850256, 9550185, 1111111111111 ]),
        )
        # From manually looking at the data
        expected = np.array([
            [ np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ],
            [ np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ],
            [ np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, ],
        ])

        npt.assert_allclose( expected, actual.values )
            
